@@ -3,7 +3,7 @@
 // FORTH compiler to generate FORTH Byte Code (FBC) from expressions
 //   or programs
 //
-// Copyright (c) 1998--2003 Krishna Myneni and David P. Wallace, 
+// Copyright (c) 1998--2004 Krishna Myneni and David P. Wallace, 
 //   Creative Consulting for Research and Education
 //
 // This software is provided under the terms of the General Public License.
@@ -83,7 +83,14 @@
 //                   for INCLUDE
 //      04-11-2003 changed F>S to FROUND>S
 //      04-15-2003 added FTRUNC and FTRUNC>S
-
+//      01-31-2004 extended intrinsic wordlist to include INCLUDE,
+//                   SOURCE, REFILL, NONDEFERRED, STATE
+//      02-09-2004 added ALLOCATE and FREE.
+//      03-18-2004 added FSINCOS, FSINH, FCOSH, FTANH, FASINH, FACOSH, 
+//                   FATANH, SP!, RP!
+//      04-xx-2004 changed intrinsic wordlist specification to
+//                   structure array following suggestion of BK.
+//      06-19-2004 updated error handling method after call to ForthVM()
 #include <iostream>
 #include <fstream>
 using std::cout;
@@ -102,7 +109,10 @@ using std::vector;
 using std::stack;
 #include "ForthCompiler.h"
 
-extern int debug;
+const int IMMEDIATE   = PRECEDENCE_IMMEDIATE;
+const int NONDEFERRED = PRECEDENCE_NON_DEFERRED;
+
+extern bool debug;
 
 // Provided by ForthVM.cpp
 
@@ -170,257 +180,283 @@ vector<byte>* pCurrentOps;
 
 DictionaryEntry NewWord;
 
-char* WordNames[] =
-    {
-        "WORD", "WORDS", "FIND", 
-	"'", "[']", "[", "]",
-	"CREATE", "DOES>", ">BODY", 
-	"FORGET", "COLD", 
-	"ALLOT", "?ALLOT",
-	"LITERAL", "EVALUATE", "IMMEDIATE",
-	"CONSTANT", "FCONSTANT",
-	"VARIABLE", "FVARIABLE",
-	"CELLS", "CELL+", "CHAR+",
-	"DFLOATS", "DFLOAT+", "SFLOATS", "SFLOAT+",
-	"?", "@", "!", 
-	"2@", "2!", "A@",
-        "C@", "C!",
-        "W@", "W!",
-        "F@", "F!",
-        "DF@", "DF!",
-        "SF@", "SF!",
-	"SP@", "RP@",
-        ">R", "R>", "R@", 
-	"2>R", "2R>", "2R@", 
-        "?DUP",
-        "DUP", "DROP", "SWAP",
-        "OVER", "ROT", "-ROT", 
-	"NIP", "TUCK", "PICK", "ROLL",
-        "2DUP", "2DROP", "2SWAP",
-        "2OVER", "2ROT",
-        "DEPTH",
-	"BASE", "BINARY", "DECIMAL", "HEX",
-        "1+", "1-", "2+", "2-", 
-	"2*", "2/",
-        "DO", "?DO", 
-	"LOOP", "+LOOP", 
-	"LEAVE", "UNLOOP", 
-	"I", "J",
-	"BEGIN", "WHILE", "REPEAT",
-	"UNTIL", "AGAIN",
-	"IF", "ELSE", "THEN",
-	"CASE", "ENDCASE", "OF", "ENDOF",
-	"RECURSE", "BYE",
-        "EXIT", "QUIT", "ABORT",
-	"ABORT\x22", "USLEEP",
-        "EXECUTE", "CALL", "SYSTEM",
-	"TIME&DATE", "MS", "MS@", 
-	"CHDIR", ">FILE", "CONSOLE",
-	"\\", "(", ".(",
-	"\x22", "C\x22", "S\x22", 
-	"COUNT", "NUMBER?",
-	"<#", "#", "#S", 
-	"#>", "SIGN", "HOLD",
-        ".", ".R",
-	"D.",
-	"U.", "U.R",
-	"F.", ".\x22", ".S",
-        "CR", "SPACES", "EMIT", "TYPE",
-	"BL", "[CHAR]", "CHAR",
-	"KEY", "KEY?", "ACCEPT",
-        "SEARCH", "COMPARE",
-        "=", "<>", "<", ">", "<=", ">=",
-	"U<", "U>",
-	"0<", "0=", "0<>", "0>",
-	"D<", "D=", "D0=",
-	"FALSE", "TRUE",
-        "AND", "OR", "XOR", "NOT", "INVERT",
-	"LSHIFT", "RSHIFT",
-        "+", "-", "*", "/",
-	"MOD", "/MOD", 
-	"*/", "*/MOD", "+!",
-	"D+", "D-",
-	"M+", "M*", "M/",
-	"M*/",
-	"UM*", "UM/MOD", 
-	"FM/MOD", "SM/REM",
-        "ABS", "NEGATE", "MIN", "MAX",
-	"DABS", "DNEGATE",
-	"OPEN", "LSEEK", "CLOSE", 
-	"READ", "WRITE", "IOCTL",
-	"FILL", "ERASE", 
-	"CMOVE", "CMOVE>",
-        "FDUP", "FDROP", "FSWAP",
-        "FOVER", "FROT",
-        "F=", "F<>", "F<", "F>", "F<=", "F>=",
-	"F0=", "F0<",
-        "F+", "F-", "F*", "F/", "F**", "FSQRT",
-        "FABS", "FNEGATE",
-	"FLOOR", "FROUND", "FTRUNC",
-	"FMIN", "FMAX",
-        "FSIN", "FCOS", "FTAN",
-        "FACOS", "FASIN", "FATAN",
-	"FATAN2",
-        "FLOG", "FLN", "FEXP",
-        "DEG>RAD", "RAD>DEG",
-        "S>D", "S>F", "D>F", "F>D",
-	"FROUND>S", "FTRUNC>S"
-    };
-
-byte WordCodes[] =
-    {
-        OP_WORD, OP_WORDS, OP_FIND, 
-	OP_TICK, OP_BRACKETTICK, OP_LBRACKET, OP_RBRACKET,
-	OP_CREATE, OP_DOES, OP_TOBODY, 
-	OP_FORGET, OP_COLD, 
-	OP_ALLOT, OP_QUERYALLOT,
-	OP_LITERAL, OP_EVALUATE, OP_IMMEDIATE,
-	OP_CONSTANT, OP_FCONSTANT,
-	OP_VARIABLE, OP_FVARIABLE,
-	OP_CELLS, OP_CELLPLUS, OP_INC,
-	OP_DFLOATS, OP_DFLOATPLUS, OP_CELLS, OP_CELLPLUS,
-	OP_QUESTION, OP_FETCH, OP_STORE, 
-	OP_DFFETCH, OP_DFSTORE, OP_AFETCH,
-        OP_CFETCH, OP_CSTORE,
-        OP_WFETCH, OP_WSTORE,
-        OP_DFFETCH, OP_DFSTORE,
-        OP_DFFETCH, OP_DFSTORE,
-        OP_SFFETCH, OP_SFSTORE,
-	OP_SPFETCH, OP_RPFETCH,
-        OP_PUSH, OP_POP, OP_RFETCH, 
-        OP_TWOPUSH, OP_TWOPOP, OP_TWORFETCH,
-        OP_QUERYDUP,
-        OP_DUP, OP_DROP, OP_SWAP,
-        OP_OVER, OP_ROT, OP_MINUSROT, 
-	OP_NIP, OP_TUCK, OP_PICK, OP_ROLL,
-        OP_2DUP, OP_2DROP, OP_2SWAP,
-        OP_2OVER, OP_2ROT,
-        OP_DEPTH,
-	OP_BASE, OP_BINARY, OP_DECIMAL, OP_HEX,
-        OP_INC, OP_DEC, OP_TWOPLUS, OP_TWOMINUS,
-	OP_TWOSTAR, OP_TWODIV,
-        OP_DO, OP_QUERYDO, 
-	OP_LOOP, OP_PLUSLOOP, 
-	OP_LEAVE, OP_UNLOOP,
-	OP_I, OP_J,
-	OP_BEGIN, OP_WHILE, OP_REPEAT,
-	OP_UNTIL, OP_AGAIN,
-	OP_IF, OP_ELSE, OP_THEN,
-	OP_CASE, OP_ENDCASE, OP_OF, OP_ENDOF,
-	OP_RECURSE, OP_BYE,
-        OP_RET, OP_QUIT, OP_ABORT,
-	OP_ABORTQUOTE, OP_USLEEP,
-        OP_EXECUTE, OP_CALL, OP_SYSTEM,
-	OP_TIMEANDDATE, OP_MS, OP_MSFETCH, 
-	OP_CHDIR, OP_TOFILE, OP_CONSOLE,
-	OP_BACKSLASH, OP_LPAREN, OP_DOTPAREN,
-	OP_CQUOTE, OP_CQUOTE, OP_SQUOTE,
-	OP_COUNT, OP_NUMBERQUERY,
-	OP_BRACKETSHARP, OP_SHARP, OP_SHARPS, 
-	OP_SHARPBRACKET, OP_SIGN, OP_HOLD,
-        OP_DOT, OP_DOTR,
-	OP_DDOT,
-	OP_UDOT, OP_UDOTR,
-	OP_FDOT, OP_DOTQUOTE, OP_DOTS,
-        OP_CR, OP_SPACES, OP_EMIT, OP_TYPE,
-	OP_BL, OP_BRACKETCHAR, OP_CHAR,
-	OP_KEY, OP_KEYQUERY, OP_ACCEPT,
-        OP_SEARCH, OP_COMPARE,
-        OP_EQ, OP_NE, OP_LT, OP_GT, OP_LE, OP_GE,
-	OP_ULT, OP_UGT,
-	OP_ZEROLT, OP_ZEROEQ, OP_ZERONE, OP_ZEROGT, 
-	OP_DLT, OP_DEQ, OP_DZEROEQ,
-	OP_FALSE, OP_TRUE,
-        OP_AND, OP_OR, OP_XOR, OP_NOT, OP_NOT,
-	OP_LSHIFT, OP_RSHIFT,
-        OP_ADD, OP_SUB, OP_MUL, OP_DIV, 
-	OP_MOD, OP_SLASHMOD,
-	OP_STARSLASH, OP_STARSLASHMOD, OP_PLUSSTORE,
-	OP_DPLUS, OP_DMINUS,
-	OP_MPLUS, OP_MSTAR, OP_MSLASH,
-	OP_MSTARSLASH,
-	OP_UMSTAR, OP_UMSLASHMOD, 
-	OP_FMSLASHMOD, OP_SMSLASHREM,
-        OP_ABS, OP_NEG, OP_MIN, OP_MAX,
-	OP_DABS, OP_DNEGATE,
-	OP_OPEN, OP_LSEEK, OP_CLOSE, 
-	OP_READ, OP_WRITE, OP_IOCTL,
-	OP_FILL, OP_ERASE, 
-	OP_CMOVE, OP_CMOVEFROM,
-        OP_2DUP, OP_2DROP, OP_2SWAP,
-        OP_2OVER, OP_2ROT,
-        OP_FEQ, OP_FNE, OP_FLT, OP_FGT, OP_FLE, OP_FGE,
-	OP_FZEROEQ, OP_FZEROLT,
-        OP_FADD, OP_FSUB, OP_FMUL, OP_FDIV, OP_FPOW, OP_FSQRT,
-        OP_FABS, OP_FNEG,
-	OP_FLOOR, OP_FROUND, OP_FTRUNC,
-	OP_FMIN, OP_FMAX,
-        OP_FSIN, OP_FCOS, OP_FTAN,
-        OP_FACOS, OP_FASIN, OP_FATAN,
-	OP_FATAN2,
-        OP_FLOG, OP_FLN, OP_FEXP,
-        OP_DEGTORAD, OP_RADTODEG,
-        OP_STOD, OP_STOF, OP_DTOF, OP_FTOD,
-	OP_FROUNDTOS, OP_FTRUNCTOS
-    };
-
-// Non-deferred words are executed immediately by
-//   the interpreter in the non-compiling state.
- 
-byte NondeferredWords[] = 
+WordTemplate IntrinsicWords[] =
 {
-  OP_BACKSLASH,
-  OP_DOTPAREN,
-  OP_BINARY, 
-  OP_DECIMAL, 
-  OP_HEX, 
-  OP_WORD, 
-  OP_TICK,
-  OP_CREATE,
-  OP_FORGET,
-  OP_COLD,
-  OP_ALLOT,
-  OP_QUERYALLOT,
-  OP_CONSTANT,
-  OP_FCONSTANT,
-  OP_VARIABLE,
-  OP_FVARIABLE,
-  OP_CHAR,
-  OP_TOFILE,
-  OP_CONSOLE,
-};
-
-byte ImmediateWords[] =
-{
-  OP_BACKSLASH,
-  OP_LPAREN,
-  OP_DOTPAREN,
-  OP_BRACKETCHAR,
-  OP_BRACKETTICK,
-  OP_LBRACKET,
-  OP_RBRACKET,
-  OP_LITERAL,
-  OP_CQUOTE,
-  OP_SQUOTE,
-  OP_DOTQUOTE,
-  OP_DO,
-  OP_QUERYDO,
-  OP_LEAVE,
-  OP_ABORTQUOTE,
-  OP_BEGIN,
-  OP_WHILE,
-  OP_REPEAT,
-  OP_UNTIL,
-  OP_AGAIN,
-  OP_IF,
-  OP_ELSE,
-  OP_THEN,
-  OP_CASE,
-  OP_ENDCASE,
-  OP_OF,
-  OP_ENDOF,
-  OP_RECURSE
+    { "WORD",      OP_WORD,         NONDEFERRED },
+    { "WORDS",     OP_WORDS,        0 },
+    { "FIND",      OP_FIND,         0 },
+    { "'",         OP_TICK,         NONDEFERRED },
+    { "[']",       OP_BRACKETTICK,  IMMEDIATE },
+    { "COMPILE,",  OP_COMPILECOMMA, 0 },
+//    { "[COMPILE]", OP_BRACKETCOMPILE, IMMEDIATE },
+    { "POSTPONE",  OP_POSTPONE,     IMMEDIATE },
+    { "[",         OP_LBRACKET,     IMMEDIATE },
+    { "]",         OP_RBRACKET,     IMMEDIATE },
+    { "STATE",     OP_STATE,        0 },
+    { "CREATE",    OP_CREATE,       NONDEFERRED },
+    { "DOES>",     OP_DOES,         0 },
+    { ">BODY",     OP_TOBODY,       0 },
+    { "FORGET",    OP_FORGET,       NONDEFERRED },
+    { "COLD",      OP_COLD,         NONDEFERRED },
+    { ":",         OP_COLON,        NONDEFERRED },
+    { ";",         OP_SEMICOLON,    IMMEDIATE },
+    { "ALLOCATE",  OP_ALLOCATE,     0 },     
+    { "FREE",      OP_FREE,         0 },
+    { "ALLOT",     OP_ALLOT,        NONDEFERRED },
+    { "?ALLOT",    OP_QUERYALLOT,   NONDEFERRED },
+    { "LITERAL",   OP_LITERAL,      IMMEDIATE },
+    { "EVALUATE",  OP_EVALUATE,     0 },
+    { "INCLUDE",   OP_INCLUDE,      NONDEFERRED },
+    { "SOURCE",    OP_SOURCE,       0 },
+    { "REFILL",    OP_REFILL,       0 }, 
+    { "IMMEDIATE", OP_IMMEDIATE,    0 },
+    { "NONDEFERRED", OP_NONDEFERRED, 0 },
+    { "CONSTANT",  OP_CONSTANT,     NONDEFERRED },
+    { "FCONSTANT", OP_FCONSTANT,    NONDEFERRED },
+    { "VARIABLE",  OP_VARIABLE,     NONDEFERRED },
+    { "FVARIABLE", OP_FVARIABLE,    NONDEFERRED },
+    { "CELLS",     OP_CELLS,        0 },
+    { "CELL+",     OP_CELLPLUS,     0 },
+    { "CHAR+",     OP_INC,          0 },
+    { "DFLOATS",   OP_DFLOATS,      0 },
+    { "DFLOAT+",   OP_DFLOATPLUS,   0 },
+    { "SFLOATS",   OP_CELLS,        0 },
+    { "SFLOAT+",   OP_CELLPLUS,     0 },
+    { "?",         OP_QUESTION,     0 },
+    { "@",         OP_FETCH,        0 },
+    { "!",         OP_STORE,        0 },
+    { "2@",        OP_DFFETCH,      0 },
+    { "2!",        OP_DFSTORE,      0 },
+    { "A@",        OP_AFETCH,       0 },
+    { "C@",        OP_CFETCH,       0 },
+    { "C!",        OP_CSTORE,       0 },
+    { "W@",        OP_WFETCH,       0 },
+    { "W!",        OP_WSTORE,       0 },
+    { "F@",        OP_DFFETCH,      0 },
+    { "F!",        OP_DFSTORE,      0 },
+    { "DF@",       OP_DFFETCH,      0 },
+    { "DF!",       OP_DFSTORE,      0 },
+    { "SF@",       OP_SFFETCH,      0 },
+    { "SF!",       OP_SFSTORE,      0 },
+    { "SP@",       OP_SPFETCH,      0 },
+    { "SP!",       OP_SPSTORE,      0 },
+    { "RP@",       OP_RPFETCH,      0 },
+    { "RP!",       OP_RPSTORE,      0 },
+    { ">R",        OP_PUSH,         0 },
+    { "R>",        OP_POP,          0 },
+    { "R@",        OP_RFETCH,       0 },
+    { "2>R",       OP_TWOPUSH,      0 }, 
+    { "2R>",       OP_TWOPOP,       0 },
+    { "2R@",       OP_TWORFETCH,    0 }, 
+    { "?DUP",      OP_QUERYDUP,     0 },
+    { "DUP",       OP_DUP,          0 },
+    { "DROP",      OP_DROP,         0 },
+    { "SWAP",      OP_SWAP,         0 },
+    { "OVER",      OP_OVER,         0 },
+    { "ROT",       OP_ROT,          0 }, 
+    { "-ROT",      OP_MINUSROT,     0 }, 
+    { "NIP",       OP_NIP,          0 },
+    { "TUCK",      OP_TUCK,         0 }, 
+    { "PICK",      OP_PICK,         0 },
+    { "ROLL",      OP_ROLL,         0 },
+    { "2DUP",      OP_2DUP,         0 },
+    { "2DROP",     OP_2DROP,        0 },
+    { "2SWAP",     OP_2SWAP,        0 },
+    { "2OVER",     OP_2OVER,        0 },
+    { "2ROT",      OP_2ROT,         0 },
+    { "DEPTH",     OP_DEPTH,        0 },
+    { "BASE",      OP_BASE,         0 },
+    { "BINARY",    OP_BINARY,       NONDEFERRED },
+    { "DECIMAL",   OP_DECIMAL,      NONDEFERRED },
+    { "HEX",       OP_HEX,          NONDEFERRED },
+    { "1+",        OP_INC,          0 },
+    { "1-",        OP_DEC,          0 },
+    { "2+",        OP_TWOPLUS,      0 },
+    { "2-",        OP_TWOMINUS,     0 },
+    { "2*",        OP_TWOSTAR,      0 },
+    { "2/",        OP_TWODIV,       0 },
+    { "DO",        OP_DO,           IMMEDIATE },
+    { "?DO",       OP_QUERYDO,      IMMEDIATE },
+    { "LOOP",      OP_LOOP,         0 },
+    { "+LOOP",     OP_PLUSLOOP,     0 },
+    { "LEAVE",     OP_LEAVE,        IMMEDIATE },
+    { "UNLOOP",    OP_UNLOOP,       0 },
+    { "I",         OP_I,            0 },
+    { "J",         OP_J,            0 },
+    { "BEGIN",     OP_BEGIN,        IMMEDIATE },
+    { "WHILE",     OP_WHILE,        IMMEDIATE },
+    { "REPEAT",    OP_REPEAT,       IMMEDIATE },
+    { "UNTIL",     OP_UNTIL,        IMMEDIATE },
+    { "AGAIN",     OP_AGAIN,        IMMEDIATE },
+    { "IF",        OP_IF,           IMMEDIATE },
+    { "ELSE",      OP_ELSE,         IMMEDIATE },
+    { "THEN",      OP_THEN,         IMMEDIATE },
+    { "CASE",      OP_CASE,         IMMEDIATE },
+    { "ENDCASE",   OP_ENDCASE,      IMMEDIATE },
+    { "OF",        OP_OF,           IMMEDIATE },
+    { "ENDOF",     OP_ENDOF,        IMMEDIATE },
+    { "RECURSE",   OP_RECURSE,      IMMEDIATE },
+    { "BYE",       OP_BYE,          0 },
+    { "EXIT",      OP_RET,          0 },
+    { "QUIT",      OP_QUIT,         0 },
+    { "ABORT",     OP_ABORT,        0 },
+    { "ABORT\x22", OP_ABORTQUOTE,   IMMEDIATE },
+    { "USLEEP",    OP_USLEEP,       0 },
+    { "EXECUTE",   OP_EXECUTE,      0 },
+    { "CALL",      OP_CALL,         0 },
+    { "SYSTEM",    OP_SYSTEM,       0 },
+    { "TIME&DATE", OP_TIMEANDDATE,  0 },
+    { "MS",        OP_MS,           0 },
+    { "MS@",       OP_MSFETCH,      0 },
+    { "CHDIR",     OP_CHDIR,        0 },
+    { ">FILE",     OP_TOFILE,       NONDEFERRED },
+    { "CONSOLE",   OP_CONSOLE,      NONDEFERRED },
+    { "\\",        OP_BACKSLASH,    IMMEDIATE | NONDEFERRED },
+    { "(",         OP_LPAREN,       IMMEDIATE },
+    { ".(",        OP_DOTPAREN,     IMMEDIATE | NONDEFERRED },
+    { "\x22",      OP_CQUOTE,       IMMEDIATE },
+    { "C\x22",     OP_CQUOTE,       IMMEDIATE },
+    { "S\x22",     OP_SQUOTE,       IMMEDIATE },
+    { "COUNT",     OP_COUNT,        0 },
+    { "NUMBER?",   OP_NUMBERQUERY,  0 },
+    { "<#",        OP_BRACKETSHARP, 0 },
+    { "#",         OP_SHARP,        0 },
+    { "#S",        OP_SHARPS,       0 },
+    { "#>",        OP_SHARPBRACKET, 0 },
+    { "SIGN",      OP_SIGN,         0 },
+    { "HOLD",      OP_HOLD,         0 },
+    { ".",         OP_DOT,          0 },
+    { ".R",        OP_DOTR,         0 },
+    { "D.",        OP_DDOT,         0 },
+    { "U.",        OP_UDOT,         0 },
+    { "U.R",       OP_UDOTR,        0 },
+    { "F.",        OP_FDOT,         0 },
+    { ".\x22",     OP_DOTQUOTE,     IMMEDIATE },
+    { ".S",        OP_DOTS,         0 },
+    { "CR",        OP_CR,           0 },
+    { "SPACES",    OP_SPACES,       0 },
+    { "EMIT",      OP_EMIT,         0 },
+    { "TYPE",      OP_TYPE,         0 },
+    { "BL",        OP_BL,           0 },
+    { "[CHAR]",    OP_BRACKETCHAR,  IMMEDIATE },
+    { "CHAR",      OP_CHAR,         NONDEFERRED },
+    { "KEY",       OP_KEY,          0 },
+    { "KEY?",      OP_KEYQUERY,     0 },
+    { "ACCEPT",    OP_ACCEPT,       0 },
+    { "SEARCH",    OP_SEARCH,       0 },
+    { "COMPARE",   OP_COMPARE,      0 },
+    { "=",         OP_EQ,           0 },
+    { "<>",        OP_NE,           0 },
+    { "<",         OP_LT,           0 },
+    { ">",         OP_GT,           0 },
+    { "<=",        OP_LE,           0 },
+    { ">=",        OP_GE,           0 },
+    { "U<",        OP_ULT,          0 },
+    { "U>",        OP_UGT,          0 },
+    { "0<",        OP_ZEROLT,       0 },
+    { "0=",        OP_ZEROEQ,       0 },
+    { "0<>",       OP_ZERONE,       0 },
+    { "0>",        OP_ZEROGT,       0 },
+    { "D<",        OP_DLT,          0 },
+    { "D=",        OP_DEQ,          0 },
+    { "D0=",       OP_DZEROEQ,      0 },
+    { "FALSE",     OP_FALSE,        0 },
+    { "TRUE",      OP_TRUE,         0 },
+    { "AND",       OP_AND,          0 },
+    { "OR",        OP_OR,           0 },
+    { "XOR",       OP_XOR,          0 },
+    { "NOT",       OP_NOT,          0 },
+    { "INVERT",    OP_NOT,          0 },
+    { "LSHIFT",    OP_LSHIFT,       0 },
+    { "RSHIFT",    OP_RSHIFT,       0 },
+    { "+",         OP_ADD,          0 },
+    { "-",         OP_SUB,          0 },
+    { "*",         OP_MUL,          0 },
+    { "/",         OP_DIV,          0 },
+    { "MOD",       OP_MOD,          0 },
+    { "/MOD",      OP_SLASHMOD,     0 },
+    { "*/",        OP_STARSLASH,    0 },
+    { "*/MOD",     OP_STARSLASHMOD, 0 },
+    { "+!",        OP_PLUSSTORE,    0 },
+    { "D+",        OP_DPLUS,        0 },
+    { "D-",        OP_DMINUS,       0 },
+    { "M+",        OP_MPLUS,        0 },
+    { "M*",        OP_MSTAR,        0 },
+    { "M/",        OP_MSLASH,       0 },
+    { "M*/",       OP_MSTARSLASH,   0 },
+    { "UM*",       OP_UMSTAR,       0 },
+    { "UM/MOD",    OP_UMSLASHMOD,   0 },
+    { "FM/MOD",    OP_FMSLASHMOD,   0 },
+    { "SM/REM",    OP_SMSLASHREM,   0 },
+    { "ABS",       OP_ABS,          0 },
+    { "NEGATE",    OP_NEG,          0 },
+    { "MIN",       OP_MIN,          0 },
+    { "MAX",       OP_MAX,          0 },
+    { "DABS",      OP_DABS,         0 },
+    { "DNEGATE",   OP_DNEGATE,      0 },
+    { "OPEN",      OP_OPEN,         0 },
+    { "LSEEK",     OP_LSEEK,        0 },
+    { "CLOSE",     OP_CLOSE,        0 },
+    { "READ",      OP_READ,         0 },
+    { "WRITE",     OP_WRITE,        0 },
+    { "IOCTL",     OP_IOCTL,        0 },
+    { "FILL",      OP_FILL,         0 },
+    { "ERASE",     OP_ERASE,        0 },
+    { "CMOVE",     OP_CMOVE,        0 },
+    { "CMOVE>",    OP_CMOVEFROM,    0 },
+    { "FDUP",      OP_2DUP,         0 },
+    { "FDROP",     OP_2DROP,        0 },
+    { "FSWAP",     OP_2SWAP,        0 },
+    { "FOVER",     OP_2OVER,        0 },
+    { "FROT",      OP_2ROT,         0 },
+    { "F=",        OP_FEQ,          0 },
+    { "F<>",       OP_FNE,          0 },
+    { "F<",        OP_FLT,          0 },
+    { "F>",        OP_FGT,          0 },
+    { "F<=",       OP_FLE,          0 },
+    { "F>=",       OP_FGE,          0 },
+    { "F0=",       OP_FZEROEQ,      0 },
+    { "F0<",       OP_FZEROLT,      0 },
+    { "F+",        OP_FADD,         0 },
+    { "F-",        OP_FSUB,         0 },
+    { "F*",        OP_FMUL,         0 },
+    { "F/",        OP_FDIV,         0 },
+    { "F**",       OP_FPOW,         0 },
+    { "FSQRT",     OP_FSQRT,        0 },
+    { "FABS",      OP_FABS,         0 },
+    { "FNEGATE",   OP_FNEG,         0 },
+    { "FLOOR",     OP_FLOOR,        0 },
+    { "FROUND",    OP_FROUND,       0 },
+    { "FTRUNC",    OP_FTRUNC,       0 },
+    { "FMIN",      OP_FMIN,         0 },
+    { "FMAX",      OP_FMAX,         0 },
+    { "FSIN",      OP_FSIN,         0 },
+    { "FCOS",      OP_FCOS,         0 },
+    { "FSINCOS",   OP_FSINCOS,      0 },
+    { "FTAN",      OP_FTAN,         0 },
+    { "FACOS",     OP_FACOS,        0 },
+    { "FASIN",     OP_FASIN,        0 },
+    { "FATAN",     OP_FATAN,        0 },
+    { "FATAN2",    OP_FATAN2,       0 },
+    { "FSINH",     OP_FSINH,        0 },
+    { "FCOSH",     OP_FCOSH,        0 },
+    { "FTANH",     OP_FTANH,        0 },
+    { "FASINH",    OP_FASINH,       0 },
+    { "FACOSH",    OP_FACOSH,       0 },
+    { "FATANH",    OP_FATANH,       0 },
+    { "FLOG",      OP_FLOG,         0 },
+    { "FLN",       OP_FLN,          0 },
+    { "FEXP",      OP_FEXP,         0 },
+    { "DEG>RAD",   OP_DEGTORAD,     0 },
+    { "RAD>DEG",   OP_RADTODEG,     0 },
+    { "S>D",       OP_STOD,         0 },
+    { "S>F",       OP_STOF,         0 },
+    { "D>F",       OP_DTOF,         0 },
+    { "F>D",       OP_FTOD,         0 },
+    { "FROUND>S",  OP_FROUNDTOS,    0 },
+    { "FTRUNC>S",  OP_FTRUNCTOS,    0 }
 };
+//---------------------------------------------------------------
 
 
 char* C_ErrorMessages[] =
@@ -439,8 +475,6 @@ char* C_ErrorMessages[] =
 	"Incomplete CASE structure",
 	"VM returned error"
 };
-
-
 //---------------------------------------------------------------
 
 char* ExtractName (char* str, char* name)
@@ -495,15 +529,12 @@ int IsFloat (char* token, double* p)
 
     char *pStr = token;
 
-//    cout << "\nIsFloat: token = " << token;
-
     if (strchr(pStr, 'E'))
     {
         while ((isdigit(*pStr)) || (*pStr == '-')
           || (*pStr == 'E') || (*pStr == '+') || (*pStr == '.'))
         {
             ++pStr;
-//            cout << ' ' << ((int) *pStr);
         }
         if (*pStr == 0)
         {
@@ -590,7 +621,7 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 {
 // The FORTH Compiler
 //
-// Reads and compile the source statements from the input stream
+// Reads and compiles the source statements from the input stream
 //   into a vector of FORTH Byte Codes.
 //
 // Return value:
@@ -644,98 +675,6 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 	  if (*pTIB == ' ' || *pTIB == '\t')
 	    ++pTIB;
 
-	  else if ((*pTIB == ':') && (*(pTIB+1) == ' '))
-	    {
-	      if (pOpCodes->size())
-		{
-		  // Execute the code outside of a definition
-
-		  pOpCodes->push_back(OP_RET);
-		  ival = ForthVM (pOpCodes, &sp, &tp);
-		  pOpCodes->erase(pOpCodes->begin(), pOpCodes->end());
-		  if (ival) 
-		    {
-		      PrintVM_Error(ival);
-		      ecode = E_C_VMERROR;
-		      goto endcompile;
-		    }
-		}
-
-	      State = TRUE;
-	      ++pTIB;
-	      pTIB = ExtractName (pTIB, WordToken);
-	      strupr(WordToken);
-	      strcpy (NewWord.WordName, WordToken);
-	      NewWord.WordCode = OP_DEFINITION;
-	      NewWord.Precedence = PRECEDENCE_NONE;
-	      NewWord.Pfa = NULL;
-	      NewWord.Cfa = NULL;
-
-	      recursestack.erase(recursestack.begin(), recursestack.end());
-	    }
-	  else if (*pTIB == ';')
-	    {
-	      pOpCodes->push_back(OP_RET);
-
-	      if (State)
-		{
-		  // Check for incomplete control structures
-		    
-		  if (ifstack.size())
-		    {
-		      ecode = E_C_INCOMPLETEIF;
-		      goto endcompile;
-		    }
-		  if (beginstack.size() || whilestack.size())
-		    {
-		      ecode = E_C_INCOMPLETEBEGIN;
-		      goto endcompile;
-		    }
-		  if (dostack.size() || leavestack.size())
-		    {
-		      ecode = E_C_INCOMPLETELOOP;
-		      goto endcompile;
-		    }
-		  if (casestack.size() || ofstack.size())
-		    {
-		      ecode = E_C_INCOMPLETECASE;
-		      goto endcompile;
-		    }
-
-		  // Add a new entry into the dictionary
-
-		  if (debug) OutputForthByteCode (pOpCodes);
- 		  
-		  NewWord.Pfa = new byte[pOpCodes->size()];
-		  NewWord.Cfa = NewWord.Pfa;
-
-		  // Resolve any self references (recursion)
-
-		  bp = (byte*) &NewWord.Pfa;
-		  while (recursestack.size())
-		    {
-		      i = recursestack[recursestack.size() - 1];
-		      ib1 = pOpCodes->begin() + i;
-		      for (i = 0; i < sizeof(void*); i++) *ib1++ = *(bp + i);
-		      recursestack.pop_back();
-		    }
-
-		  byte* dest = (byte*) NewWord.Pfa;
-		  bp = (byte*) &(*pOpCodes)[0]; // ->begin();
-		  while ((vector<byte>::iterator) bp < pOpCodes->end()) *dest++ = *bp++;
-		  if (IsForthWord(NewWord.WordName, &d))
-		    *pOutStream << NewWord.WordName << " is redefined\n";
-		  Dictionary.push_back(NewWord);
-		  pOpCodes->erase(pOpCodes->begin(), pOpCodes->end());
-		  State = FALSE;
-		}
-	      else
-		{
-		  ecode = E_C_ENDOFDEF;
-		  goto endcompile;
-		}
-	      ++pTIB;
-	    }
 	  else
 	    {
 	      pTIB = ExtractName (pTIB, WordToken);
@@ -743,8 +682,8 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 
 	      if (IsForthWord(WordToken, &d))
 		{
-		  pOpCodes->push_back(d.WordCode);
-		  
+		    if (d.WordCode >> 8) d.WordCode = OP_DEFINITION;
+		    pOpCodes->push_back(d.WordCode);		  
 
 		  if (d.WordCode == OP_DEFINITION)
 		    {
@@ -821,49 +760,25 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 
 		  int execution_method = EXECUTE_NONE;
 
-		  if (postpone)
+		  switch (d.Precedence)
 		    {
-		      if ((d.Precedence & PRECEDENCE_IMMEDIATE) == 0)
-			{
-			  id = LocateWord (WordToken);
-			  i = (d.WordCode == OP_DEFINITION) ? 5 : 1;
-			  ib1 = pOpCodes->end() - i;
-			  pOpCodes->erase(ib1, pOpCodes->end());
-			  i = strlen(id->WordName);;
-			  str = new char[i + 1];
-			  strcpy(str, id->WordName);
-			  pOpCodes->push_back(OP_ADDR);
-			  StringTable.push_back(str);
-			  OpsPushInt((int) str);
-			  pOpCodes->push_back(OP_IVAL);
-			  OpsPushInt(i);
-			  pOpCodes->push_back(OP_EVALUATE);
-			}
-		      postpone = FALSE;
-		      if (State && (d.Precedence | PRECEDENCE_NON_DEFERRED))
-			NewWord.Precedence |= PRECEDENCE_NON_DEFERRED;
-		    }
-		  else
-		    {
-		      switch (d.Precedence)
-			{
-			case PRECEDENCE_IMMEDIATE:
-			  execution_method = EXECUTE_CURRENT_ONLY;
-			  break;
-			case PRECEDENCE_NON_DEFERRED:
-			  if (State)
-			    NewWord.Precedence |= PRECEDENCE_NON_DEFERRED ;
-			  else
-			    execution_method = EXECUTE_UP_TO;
-			  break;
-			case (PRECEDENCE_NON_DEFERRED + PRECEDENCE_IMMEDIATE):
-			  execution_method = State ? EXECUTE_CURRENT_ONLY :
-			    EXECUTE_UP_TO;
-			  break;
-			default:
-			  ;
-			}
+		      case IMMEDIATE:
+			execution_method = EXECUTE_CURRENT_ONLY;
+			break;
+		      case NONDEFERRED:
+			if (State)
+			  NewWord.Precedence |= NONDEFERRED ;
+			else
+			  execution_method = EXECUTE_UP_TO;
+			break;
+		      case (NONDEFERRED + IMMEDIATE):
+			execution_method = State ? EXECUTE_CURRENT_ONLY :
+			  EXECUTE_UP_TO;
+			break;
+		      default:
+			;
 		    }
+
 		  vector<byte> SingleOp;
 		  
 		  switch (execution_method)
@@ -874,13 +789,9 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 
 		      pOpCodes->push_back(OP_RET);
 		      if (debug) OutputForthByteCode (pOpCodes);
-		      ival = ForthVM (pOpCodes, &sp, &tp);
+		      ecode = ForthVM (pOpCodes, &sp, &tp);
 		      pOpCodes->erase(pOpCodes->begin(), pOpCodes->end());
-		      if (ival) 
-			{ 
-			  PrintVM_Error(ival); ecode = E_C_VMERROR;
-			  goto endcompile; 
-			}
+		      if (ecode) goto endcompile; 
 		      break;
 
 		    case EXECUTE_CURRENT_ONLY:
@@ -889,13 +800,9 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 		      for (j = 0; j < i; j++) SingleOp.push_back(*(ib1+j));
 		      SingleOp.push_back(OP_RET);
 		      pOpCodes->erase(ib1, pOpCodes->end());
-		      ival = ForthVM (&SingleOp, &sp, &tp);
+		      ecode = ForthVM (&SingleOp, &sp, &tp);
 		      SingleOp.erase(SingleOp.begin(), SingleOp.end());
-		      if (ival) 
-			{ 
-			  PrintVM_Error(ival); ecode = E_C_VMERROR;
-			  goto endcompile; 
-			}
+		      if (ecode) goto endcompile; 
 		      pOpCodes = pCurrentOps; // may have been redirected
 		      break;
 
@@ -916,93 +823,6 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 		  for (i = 0; i < sizeof(double); i++)
 		    pOpCodes->push_back(*(fp + i)); // store in proper order
 		}
-	      else if (strcmp(WordToken, "STATE") == 0)
-		{
-		  pOpCodes->push_back(OP_ADDR);
-		  OpsPushInt((int) &State);
-		}
-	      else if (strcmp(WordToken, "POSTPONE") == 0)
-		{
-		  postpone = TRUE;
-		}
-	      else if (strcmp(WordToken, "NONDEFERRED") == 0)
-		{
-		  CPP_nondeferred();
-		}
-	      else if (strcmp(WordToken, "SOURCE") == 0)
-	        {
-		    pOpCodes->push_back(OP_ADDR);
-		    OpsPushInt((int) CPP_source);
-		    pOpCodes->push_back(OP_CALL);
-		}
-	      else if (strcmp(WordToken, "REFILL") == 0)
-	        {
-		    pOpCodes->push_back(OP_ADDR);
-		    OpsPushInt((int) CPP_refill);
-		    pOpCodes->push_back(OP_CALL);
-		}
-	      else if (strcmp(WordToken, "INCLUDE") == 0)
-		{
-		  if (State)
-		    {
-		      ecode = E_C_NOTINDEF;
-		      goto endcompile;
-		    }
-		  ++pTIB;
-		  pTIB = ExtractName (pTIB, WordToken);
-		  strcpy (s, pTIB);  // save remaining part of input line in TIB
-		  if (!strchr(WordToken, '.')) strcat(WordToken, ".4th");
-		  strcpy (filename, WordToken);
-		  ifstream f(filename);
-		  if (!f)
-		  {
-		      if(!strchr(filename,'/'))
-		      {
-			char temp[256];
-			if (getenv("KFORTH_DIR"))
-			  strcpy(temp, getenv("KFORTH_DIR"));
-			strcat(temp, "/");
-			strcat(temp, filename);
-			strcpy(filename, temp);
-			f.open(filename);
-			if (f) 
-			  {
-			    *pOutStream << endl << filename << endl;
-			  }
-		      }
-		  }
-
-		  if (f.fail()) 
-		    {
-		      *pOutStream << endl << filename << endl;
-		      ecode = E_C_OPENFILE;
-		      goto endcompile;
-		    }
-		  istream* pTempIn = pInStream;  // save input stream ptr
-		  SetForthInputStream(f);  // set the new input stream
-		  int oldlc = linecount; linecount = 0;
-		  ecode = ForthCompiler (pOpCodes, &linecount);
-		  f.close();
-		  pInStream = pTempIn;  // restore the input stream
-		  if (ecode) 
-		    {
-		      *pOutStream << filename << "  " ;
-		      goto endcompile;
-		    }
-		  linecount = oldlc;
-
-		  // Execute the code immediately
-		      
-		  ival = ForthVM (pOpCodes, &sp, &tp);
-		  pOpCodes->erase(pOpCodes->begin(), pOpCodes->end());
-		  if (ival) 
-		    { 
-		      PrintVM_Error(ival); ecode = E_C_VMERROR; 
-		      goto endcompile; 
-		    }
-		  strcpy(TIB, s);  // restore TIB with remaining input line
-		  pTIB = TIB;      // restore ptr
-		}
 	      else
 		{
 		  *pOutStream << endl << WordToken << endl;
@@ -1017,12 +837,9 @@ int ForthCompiler (vector<byte>* pOpCodes, int* pLc)
 	  // Execute the current line in interpretation state
 	  pOpCodes->push_back(OP_RET);
 	  if (debug) OutputForthByteCode (pOpCodes);
-	  ival = ForthVM (pOpCodes, &sp, &tp);
+	  ecode = ForthVM (pOpCodes, &sp, &tp);
 	  pOpCodes->erase(pOpCodes->begin(), pOpCodes->end());
-	  if (ival) 
-	    { 
-	      PrintVM_Error(ival); ecode = E_C_VMERROR; goto endcompile; 
-	    }
+	  if (ecode) goto endcompile; 
 	}
 
     } // end while (TRUE)
@@ -1031,7 +848,7 @@ endcompile:
     
   if ((ecode != E_C_NOERROR) && (ecode != E_C_ENDOFSTREAM))
     {
-      // A compiler error occurred; reset to interpreter mode and
+      // A compiler or vm error occurred; reset to interpreter mode and
       //   clear all flow control stacks.
 
       State = FALSE;
