@@ -2,7 +2,7 @@
 //
 // The FORTH Virtual Machine (FVM) to execute Forth byte code.
 //
-// Copyright (c) 1996--2003 Krishna Myneni, Creative Consulting for
+// Copyright (c) 1996--2004 Krishna Myneni, Creative Consulting for
 //   Research & Education
 //
 // This software is provided under the General Public License.
@@ -60,6 +60,26 @@
 //      09-29-2002  fixed CPP_word to remove delimiter from input stream  km
 //      04-11-2003  fixed problems with CPP_word and added CPP_source and
 //                    CPP_refill  km
+//      01-31-2004  added CPP_include, CPP_state, CPP_nondeferred  km
+//      02-09-2004  added CPP_allocate and CPP_free  km
+//      03-17-2004  modified OpenForth() to initialize from WordTemplate array km
+//      03-18-2004  added CPP_spstore  km
+//      03-27-2004  changed LocateWord to pre-increment iterator as suggested
+//                    by BK; improves efficiency by 24%; also use 2-byte
+//                    precomparison to gain another 12% improvement  km
+//      03-29-2004  replaced calls to L_2dup and L_2drop with C++ code so
+//                    that these words can use NEXT in vm.s  km
+//      04-08-2004  added CPP_postpone(), CPP_compilecomma(), and skeleton for
+//                    CPP_bracketcompile(), and auxiliary words OpsCompileByte(),
+//                    and OpsCompileInt()  km
+//      04-09-2004  revised CPP_postpone() to handle variables, constants, and
+//                    2constants. Added auxiliary word OpsCompileDouble().
+//                    Added CPP_semicolon()  km
+//      04-23-2004  changed variable "debug" to type "bool"  km
+//      04-26-2004  fixed problem with CPP_sharp() -- needed to use unsigned 
+//                    arithmetic. Import L_udmstar and L_utmslash from vm.s.
+//                    Now passes John Hayes' ANS core tests for <# # #S #>  km 
+//      06-19-2004  modified PrintVM_Error() to handle both kinds of errors km
 #include <string.h>
 #include <stdlib.h>
 #include <math.h>
@@ -82,17 +102,15 @@ using std::vector;
 #define STACK_SIZE 32768
 #define RETURN_STACK_SIZE 4096
 
-extern int debug;
+extern bool debug;
 
 // Provided by ForthCompiler.cpp
 
 void strupr (char*);
 void SetForthInputStream (istream&);
 void SetForthOutputStream (ostream&);
-extern char* WordNames[];
-extern byte WordCodes[];
-extern byte NondeferredWords[];
-extern byte ImmediateWords[];
+extern WordTemplate IntrinsicWords[];
+extern char* C_ErrorMessages[];
 extern int linecount;
 extern istream* pInStream ;    // global input stream
 extern ostream* pOutStream ;   // global output stream
@@ -121,10 +139,10 @@ extern "C" {
   int L_abort();
   int L_ret();
   int L_dabs();
-  int L_2dup();
-  int L_2drop();
   int L_dminus();
   int L_mstarslash();
+  int L_udmstar();
+  int L_utmslash();
   int vm (byte*);     // the machine code virtual machine
 
 
@@ -149,6 +167,7 @@ extern "C" {
   char NumberBuf[256];
 }
 
+extern "C" int JumpTable[];
 
 // The Dictionary
 
@@ -188,7 +207,9 @@ char* V_ErrorMessages[] =
 	"ELSE without matching IF",
 	"THEN without matching IF",
 	"ENDOF without matching OF",
-	"ENDCASE without matching CASE"
+	"ENDCASE without matching CASE",
+	"Cannot open file",
+	"Address outside of stack space"
 };
 
 
@@ -199,50 +220,36 @@ int OpenForth ()
 // Initialize the FORTH dictionary; return the size of
 //   the dictionary.
 
-    int i;
+    int i, wcode;
     DictionaryEntry d;
 
     set_start_time();
 
     for (i = 0; i < NUMBER_OF_INTRINSIC_WORDS; i++)
     {
-        strcpy(d.WordName, WordNames[i]);
-        d.WordCode = WordCodes[i];
-	d.Precedence = 0;
-        d.Pfa = new byte[2];
+        strcpy(d.WordName, IntrinsicWords[i].WordName);
+        wcode = IntrinsicWords[i].WordCode;
+	d.WordCode = wcode;
+	d.Precedence = IntrinsicWords[i].Precedence;
+        d.Pfa = new byte[8];
 	d.Cfa = d.Pfa;
 	byte* bp = (byte*) d.Pfa;
-	bp[0] = d.WordCode;
-	bp[1] = OP_RET;
+	if (wcode >> 8)
+	{
+	    bp[0] = OP_ADDR;
+	    *((int*) (bp+1)) = (int) JumpTable[wcode];
+	    bp[5] = OP_CALL;
+	    bp[6] = OP_RET;
+	}
+	else
+	{
+	    bp[0] = wcode;
+	    bp[1] = OP_RET;
+	}
 	
         Dictionary.push_back(d);
     }
 
-    // Set up precedence for immediate and non-deferred words
-	
-    vector<DictionaryEntry>::iterator wI;
-
-    for (wI = Dictionary.begin(); wI < Dictionary.end(); wI++)
-      {
-	for (i = 0; i < NUMBER_OF_IMMEDIATE_WORDS; i++)
-	  {
-	    if (wI->WordCode == ImmediateWords[i])
-	      {
-		wI->Precedence |= PRECEDENCE_IMMEDIATE;
-		break;
-	      }
-	  }
-	for (i = 0; i < NUMBER_OF_NON_DEFERRED_WORDS; i++)
-	  {
-	    if (wI->WordCode == NondeferredWords[i])
-	      {
-		wI->Precedence |= PRECEDENCE_NON_DEFERRED;
-		break;
-	      }
-	  }
-      }
-
-
     // Initialize the global stack pointers
 
     BottomOfStack = ForthStack + STACK_SIZE - 1;
@@ -306,8 +313,9 @@ vector<DictionaryEntry>::iterator LocateWord (char* name)
 
 	vector<DictionaryEntry>::iterator i;
     
-	for (i = Dictionary.end()-1; i >= Dictionary.begin(); i--)
+	for (i = Dictionary.end()-1; i >= Dictionary.begin(); --i)
 	{
+	    if (*((word*) name) == *((word*) i->WordName)) // pre-compare
         	if (strcmp(name, i->WordName) == 0) break;
 	}
 
@@ -354,11 +362,52 @@ void OpsPushInt (int i)
 }
 //---------------------------------------------------------------
 
-void PrintVM_Error (int ecode)
+int OpsCompileByte ()
 {
-  if ((ecode >= 0) && (ecode < MAX_V_ERR_MESSAGES))
-    *pOutStream << "VM Error(" << ecode << "): " <<
-      V_ErrorMessages[ecode] << '\n';
+  // push a byte value from the stack into the current opcode vector
+
+  ++GlobalSp; ++GlobalTp;
+  byte* ip = (byte*) GlobalSp;
+  pCurrentOps->push_back(*ip);
+  return 0;
+}
+//---------------------------------------------------------------
+
+int OpsCompileInt ()
+{
+  // push an int value from the stack into the current opcode vector
+
+  ++GlobalSp; ++GlobalTp;
+  OpsPushInt(*GlobalSp);
+  return 0;
+}
+//---------------------------------------------------------------
+
+int OpsCompileDouble ()
+{
+  // push a double value from the stack into the current opcode vector
+
+  ++GlobalSp; ++GlobalTp;
+  OpsPushInt(*GlobalSp++); ++GlobalTp;
+  OpsPushInt(*GlobalSp);
+  return 0;
+}
+//----------------------------------------------------------------
+
+void PrintVM_Error (int ec)
+{
+    int ei = ec & 0xFF;
+    int imax = (ec >> 8) ? MAX_C_ERR_MESSAGES : MAX_V_ERR_MESSAGES;
+    char *pMsg, elabel[12];
+    
+    if ((ei >= 0) * (ei < imax))
+    {
+	pMsg = (ec >> 8) ? C_ErrorMessages[ei] : V_ErrorMessages[ei];
+	if (ec >> 8)  strcpy(elabel, "Compiler"); 
+	else strcpy(elabel, "VM");
+	*pOutStream << elabel << " Error(" << ei << "): " <<
+	    pMsg << endl;
+   }
 }
 //---------------------------------------------------------------
 
@@ -427,6 +476,91 @@ if (debug)  cout << "<ForthVM Sp: " << GlobalSp << " Rp: " << GlobalRp <<
 
 extern "C" {
 
+int CPP_colon()
+{
+    // stack: ( -- | the colon compiler )
+
+    char WordToken[256];
+    State = TRUE;
+    ++pTIB;
+    pTIB = ExtractName (pTIB, WordToken);
+    strupr(WordToken);
+    strcpy (NewWord.WordName, WordToken);
+    NewWord.WordCode = OP_DEFINITION;
+    NewWord.Precedence = PRECEDENCE_NONE;
+    NewWord.Pfa = NULL;
+    NewWord.Cfa = NULL;
+    recursestack.erase(recursestack.begin(), recursestack.end());
+
+    return 0;
+}
+//----------------------------------------------------------------
+
+int CPP_semicolon()
+{
+  // stack: ( -- | terminate compilation of word )
+
+  int ecode = 0;
+
+  pCurrentOps->push_back(OP_RET);
+
+  if (State)
+    {
+      // Check for incomplete control structures
+		    
+      if (ifstack.size())                          ecode = E_C_INCOMPLETEIF;
+
+      if (beginstack.size() || whilestack.size())  ecode = E_C_INCOMPLETEBEGIN;
+      
+      if (dostack.size() || leavestack.size())     ecode = E_C_INCOMPLETELOOP;
+
+      if (casestack.size() || ofstack.size())      ecode = E_C_INCOMPLETECASE;
+      
+      if (ecode) return ecode;
+
+      // Add a new entry into the dictionary
+
+      if (debug) OutputForthByteCode (pCurrentOps);
+ 		  
+      NewWord.Pfa = new byte[pCurrentOps->size()];
+      NewWord.Cfa = NewWord.Pfa;
+
+      // Resolve any self references (recursion)
+
+      byte *bp, *dest;
+      int i;
+      vector<byte>::iterator ib;
+      DictionaryEntry d;
+
+
+      bp = (byte*) &NewWord.Pfa;
+      while (recursestack.size())
+	{
+	  i = recursestack[recursestack.size() - 1];
+	  ib = pCurrentOps->begin() + i;
+	  for (i = 0; i < sizeof(void*); i++) *ib++ = *(bp + i);
+	  recursestack.pop_back();
+	}
+
+      dest = (byte*) NewWord.Pfa;
+      bp = (byte*) &(*pCurrentOps)[0]; // ->begin();
+      while ((vector<byte>::iterator) bp < pCurrentOps->end()) *dest++ = *bp++;
+      if (IsForthWord(NewWord.WordName, &d))
+	*pOutStream << NewWord.WordName << " is redefined\n";
+      Dictionary.push_back(NewWord);
+      pCurrentOps->erase(pCurrentOps->begin(), pCurrentOps->end());
+      State = FALSE;
+    }
+  else
+    {
+      ecode = E_C_ENDOFDEF;
+      // goto endcompile;
+    }
+    
+  return ecode;
+}
+//-----------------------------------------------------------------
+
 int CPP_backslash()
 {
   // stack: ( -- | advance pTIB to end of line )
@@ -512,18 +646,22 @@ int CPP_sharp()
   unsigned int u1, u2, rem;
   char ch;
 
-  L_2dup();
-  *GlobalSp-- = 1; --GlobalTp;
+  *GlobalSp-- = *(GlobalSp+2); *GlobalSp-- = *(GlobalSp+2); // 2dup
+  *GlobalTp-- = *(GlobalTp+2); *GlobalTp-- = *(GlobalTp+2); //  "
+  *GlobalSp-- = 0; --GlobalTp;  // pad to triple length
   *GlobalSp-- = Base; --GlobalTp;
-  L_mstarslash();
+  L_utmslash();
   u1 = *(GlobalSp + 1);  // quotient
   u2 = *(GlobalSp + 2);
+
   // quotient is on the stack; we need the remainder
+
   *GlobalSp-- = Base; --GlobalTp;
-  *GlobalSp-- = 1; --GlobalTp;
-  L_mstarslash();
+  L_udmstar();
+  ++GlobalSp; ++GlobalTp;  // drop
   L_dminus();
   rem = *(GlobalSp + 2);  // get the remainder
+
   *(GlobalSp + 1) = u1;   // replace rem with quotient on the stack
   *(GlobalSp + 2) = u2;
   ch = (rem < 10) ? (rem + 48) : (rem + 55);
@@ -578,7 +716,7 @@ int CPP_sharpbracket()
 {
   // stack: ( 0 0 -- | complete number conversion )
 
-  L_2drop();
+  ++GlobalSp; ++GlobalSp; ++GlobalTp; ++GlobalTp; // 2drop
   *GlobalSp-- = (int) (NumberBuf + 255 - NumberCount);
   *GlobalTp-- = OP_ADDR;
   *GlobalSp-- = NumberCount;
@@ -944,6 +1082,36 @@ int CPP_words ()
 }
 //---------------------------------------------------------------
 
+int CPP_allocate()
+{
+    // stack: ( u -- a ior | allocate u bytes and return address and success)
+  ++GlobalSp; ++GlobalTp;
+  if (GlobalSp > BottomOfStack) 
+    return E_V_STK_UNDERFLOW;
+  if (*GlobalTp != OP_IVAL)
+    return E_V_NOTIVAL;  // need an int
+
+  unsigned int requested = *GlobalSp;
+  byte *p = new byte[requested];
+  *GlobalSp-- = (int) p; *GlobalTp-- = OP_ADDR;
+  *GlobalSp-- = p ? 0 : -1; *GlobalTp-- = OP_IVAL;
+
+  return 0;
+}
+//----------------------------------------------------------------
+
+int CPP_free()
+{
+    // stack: ( a -- ior | free the allocated region at address a )
+    ++GlobalSp; ++GlobalTp;
+    if (*GlobalTp != OP_ADDR) return E_V_NOTADDR; // need an addr
+    byte *p = (byte*) *GlobalSp; 
+    delete [] p;
+    *GlobalSp-- = 0; *GlobalTp-- = OP_IVAL;
+    return 0;
+}
+//----------------------------------------------------------------
+
 int CPP_allot ()
 {
   ++GlobalSp; ++GlobalTp;
@@ -1165,6 +1333,89 @@ int CPP_brackettick ()
 }
 //-------------------------------------------------------------------
 
+int CPP_compilecomma ()
+{
+  // stack: xt --
+  CPP_literal();
+  pCurrentOps->push_back(OP_EXECUTE);
+  return 0;  
+}
+// -----------------------------------------------------------------
+
+int CPP_bracketcompile ()
+{
+  return 0;
+}
+// -----------------------------------------------------------------
+
+int CPP_postpone ()
+{
+  char token[128];
+  ++pTIB;
+  pTIB = ExtractName (pTIB, token);
+  strupr(token);
+
+  vector<DictionaryEntry>::iterator id = LocateWord(token);
+  if (id != (vector<DictionaryEntry>::iterator) NULL)
+    {
+      if (id->Precedence & PRECEDENCE_IMMEDIATE)
+	{
+	  pCurrentOps->push_back(id->WordCode);
+	  if (id->WordCode == OP_DEFINITION)
+	    OpsPushInt((int) id->Cfa);
+	}
+      else
+	{
+	  if (id->WordCode == OP_IVAL)
+	    {
+	      pCurrentOps->push_back(OP_IVAL);
+	      OpsPushInt (*((int*) id->Pfa));
+	      pCurrentOps->push_back(OP_LITERAL);
+	    }
+	  else if (id->WordCode == OP_ADDR)
+	    {
+	      pCurrentOps->push_back(OP_ADDR);
+	      OpsPushInt ((int) id->Pfa);
+	      pCurrentOps->push_back(OP_LITERAL);
+	    }
+	  else
+	    {
+	      pCurrentOps->push_back(OP_IVAL);
+	      OpsPushInt((int) id->WordCode);
+	      pCurrentOps->push_back(OP_ADDR);
+	      OpsPushInt((int) OpsCompileByte);
+	      pCurrentOps->push_back(OP_CALL);
+	      if (id->WordCode == OP_DEFINITION)
+		{
+		  pCurrentOps->push_back(OP_ADDR);
+		  OpsPushInt((int) id->Cfa);
+		  pCurrentOps->push_back(OP_ADDR);
+		  OpsPushInt((int) OpsCompileInt);
+		  pCurrentOps->push_back(OP_CALL);
+		}
+	      else if (id->WordCode == OP_FVAL)
+		{
+	          pCurrentOps->push_back(OP_FVAL);
+	          int* pVal = (int*) id->Pfa;
+	          OpsPushInt (*pVal);
+	          OpsPushInt (*(pVal+1));
+	          pCurrentOps->push_back(OP_ADDR);
+	          OpsPushInt((int) OpsCompileDouble);
+	          pCurrentOps->push_back(OP_CALL);
+		}	  
+	    } 
+	}
+      if (State && (id->Precedence & PRECEDENCE_NON_DEFERRED)) 
+	NewWord.Precedence |= PRECEDENCE_NON_DEFERRED;
+
+    }
+
+
+  return 0;
+  
+}
+// -----------------------------------------------------------------
+
 int CPP_forget ()
 {
   char token[128];
@@ -1724,6 +1975,74 @@ int CPP_evaluate ()
 }
 //-------------------------------------------------------------------
 
+int CPP_include()
+{
+    char WordToken[256], filename[256], s[256];
+    int ecode;
+    vector<byte> ops, *pOldOps;
+
+    ++pTIB;
+    pTIB = ExtractName (pTIB, WordToken);
+    strcpy (s, pTIB);  // save remaining part of input line in TIB
+    if (!strchr(WordToken, '.')) strcat(WordToken, ".4th");
+    strcpy (filename, WordToken);
+    ifstream f(filename);
+    if (!f)
+    {
+	if(!strchr(filename,'/'))
+	{
+	    char temp[256];
+	    if (getenv("KFORTH_DIR"))
+	    { 
+		strcpy(temp, getenv("KFORTH_DIR"));
+	        strcat(temp, "/");
+	        strcat(temp, filename);
+		strcpy(filename, temp);
+		f.open(filename);
+		if (f) 
+		{
+		    *pOutStream << endl << filename << endl;
+		}
+	    }
+	}
+    }
+
+    if (f.fail()) 
+    {
+	*pOutStream << endl << filename << endl;
+	return (E_V_OPENFILE);
+    }
+	
+    istream* pTempIn = pInStream;  // save input stream ptr
+    SetForthInputStream(f);  // set the new input stream
+    int oldlc = linecount; linecount = 0;
+    pOldOps = pCurrentOps;
+    ecode = ForthCompiler (&ops, &linecount);
+    f.close();
+    pInStream = pTempIn;  // restore the input stream
+    pCurrentOps = pOldOps; 
+    if (ecode) 
+    {
+	*pOutStream << filename << "  " ;
+	return (ecode);
+    }
+    linecount = oldlc;
+
+    // Execute the code immediately
+		      
+    int *sp;
+    byte *tp;
+    ecode = ForthVM (&ops, &sp, &tp);
+    ops.erase(ops.begin(), ops.end());
+    if (ecode) return(ecode);
+
+    strcpy(TIB, s);  // restore TIB with remaining input line
+    pTIB = TIB;      // restore ptr
+    
+    return 0;
+}
+//-------------------------------------------------------------------
+
 int CPP_source()
 {
     *GlobalSp-- = (int) TIB;
@@ -1744,12 +2063,49 @@ int CPP_refill()
 }
 //-------------------------------------------------------------------
 
-int CPP_dump ()
+int CPP_state()
 {
-  // stack: ( a u -- | display memory; u bytes starting at address a )
+    // pCurrentOps->push_back(OP_ADDR);
+    // OpsPushInt((int) &State);
+    *GlobalSp-- = (int)(&State);
+    *GlobalTp-- = OP_ADDR;
+    return 0;
+}
+//-------------------------------------------------------------------
 
-  return 0;
+int CPP_spstore()
+{
+    // stack: ( addr -- | make the stack ptr point to a new address)
+
+    ++GlobalSp; ++GlobalTp;
+    if (*GlobalTp != OP_ADDR) return E_V_NOTADDR;  // need an address
+    int* p = (int*)(*GlobalSp); --p;
+    if ((p > BottomOfStack) || (p < ForthStack))
+	return E_V_BADSTACKADDR;  // new SP must be within its stack space
+
+    int n = (int) (p - ForthStack);
+    GlobalSp = ForthStack + n;
+    GlobalTp = ForthTypeStack + n;
+
+    return 0;
 }
 //--------------------------------------------------------------------
 
+int CPP_rpstore()
+{
+    // stack: ( addr -- | make the stack ptr point to a new address)
+
+    ++GlobalSp; ++GlobalTp;
+    if (*GlobalTp != OP_ADDR) return E_V_NOTADDR;  // need an address
+    int* p = (int*)(*GlobalSp); --p;
+    if ((p > BottomOfReturnStack) || (p < ForthReturnStack))
+	return E_V_BADSTACKADDR;  // new RP must be within its stack space
+
+    int n = (int) (p - ForthReturnStack);
+    GlobalRp = ForthReturnStack + n;
+    GlobalRtp = ForthReturnTypeStack + n;
+
+    return 0;
+}
+
 }
